@@ -15,6 +15,7 @@ cfg_if::cfg_if! {
   }
 }
 
+use crate::api::SGRComplexityLevel;
 use crate::color::ChromaSampling::Cs400;
 use crate::context::{MAX_PLANES, SB_SIZE};
 use crate::encoder::FrameInvariants;
@@ -24,8 +25,6 @@ use crate::frame::{
 use crate::tiling::{Area, PlaneRegion, PlaneRegionMut, Rect};
 use crate::util::{clamp, CastFromPrimitive, ILog, Pixel};
 use rust_hawktracer::*;
-
-use crate::api::SGRComplexityLevel;
 use std::cmp;
 use std::iter::FusedIterator;
 use std::ops::{Index, IndexMut};
@@ -174,32 +173,39 @@ pub(crate) mod rust {
   use crate::Pixel;
 
   #[inline(always)]
-  pub(crate) fn sgrproj_box_ab_internal(
+  pub(crate) fn sgrproj_box_ab_internal<const BD: usize>(
     r: usize, af: &mut [u32], bf: &mut [u32], iimg: &[u32], iimg_sq: &[u32],
     iimg_stride: usize, start_x: usize, y: usize, stripe_w: usize, s: u32,
-    bdm8: usize,
   ) {
     let d: usize = r * 2 + 1;
     let n: usize = d * d;
     let one_over_n = if r == 1 { 455 } else { 164 };
 
+    assert!(iimg.len() > (y + d) * iimg_stride + stripe_w + 1 + d);
+    assert!(iimg_sq.len() > (y + d) * iimg_stride + stripe_w + 1 + d);
+    assert!(af.len() > stripe_w + 1);
+    assert!(bf.len() > stripe_w + 1);
+
     for x in start_x..stripe_w + 2 {
-      let sum = get_integral_square(iimg, iimg_stride, x, y, d);
-      let ssq = get_integral_square(iimg_sq, iimg_stride, x, y, d);
-      let (reta, retb) =
-        sgrproj_sum_finish(ssq, sum, n as u32, one_over_n, s, bdm8);
-      af[x] = reta;
-      bf[x] = retb;
+      // SAFETY: We perform the bounds checks above, once for the whole loop
+      unsafe {
+        let sum = get_integral_square(iimg, iimg_stride, x, y, d);
+        let ssq = get_integral_square(iimg_sq, iimg_stride, x, y, d);
+        let (reta, retb) =
+          sgrproj_sum_finish::<BD>(ssq, sum, n as u32, one_over_n, s);
+        *af.get_unchecked_mut(x) = reta;
+        *bf.get_unchecked_mut(x) = retb;
+      }
     }
   }
 
   // computes an intermediate (ab) row for stripe_w + 2 columns at row y
-  pub(crate) fn sgrproj_box_ab_r1(
+  pub(crate) fn sgrproj_box_ab_r1<const BD: usize>(
     af: &mut [u32], bf: &mut [u32], iimg: &[u32], iimg_sq: &[u32],
-    iimg_stride: usize, y: usize, stripe_w: usize, s: u32, bdm8: usize,
+    iimg_stride: usize, y: usize, stripe_w: usize, s: u32,
     _cpu: CpuFeatureLevel,
   ) {
-    sgrproj_box_ab_internal(
+    sgrproj_box_ab_internal::<BD>(
       1,
       af,
       bf,
@@ -210,17 +216,16 @@ pub(crate) mod rust {
       y,
       stripe_w,
       s,
-      bdm8,
     );
   }
 
   // computes an intermediate (ab) row for stripe_w + 2 columns at row y
-  pub(crate) fn sgrproj_box_ab_r2(
+  pub(crate) fn sgrproj_box_ab_r2<const BD: usize>(
     af: &mut [u32], bf: &mut [u32], iimg: &[u32], iimg_sq: &[u32],
-    iimg_stride: usize, y: usize, stripe_w: usize, s: u32, bdm8: usize,
+    iimg_stride: usize, y: usize, stripe_w: usize, s: u32,
     _cpu: CpuFeatureLevel,
   ) {
-    sgrproj_box_ab_internal(
+    sgrproj_box_ab_internal::<BD>(
       2,
       af,
       bf,
@@ -231,7 +236,6 @@ pub(crate) mod rust {
       y,
       stripe_w,
       s,
-      bdm8,
     );
   }
 
@@ -338,14 +342,13 @@ pub(crate) mod rust {
 }
 
 #[inline(always)]
-fn sgrproj_sum_finish(
-  ssq: u32, sum: u32, n: u32, one_over_n: u32, s: u32, bdm8: usize,
+fn sgrproj_sum_finish<const BD: usize>(
+  ssq: u32, sum: u32, n: u32, one_over_n: u32, s: u32,
 ) -> (u32, u32) {
+  let bdm8 = BD - 8;
   let scaled_ssq = (ssq + (1 << (2 * bdm8) >> 1)) >> (2 * bdm8);
   let scaled_sum = (sum + (1 << bdm8 >> 1)) >> bdm8;
-  let p =
-    cmp::max(0, (scaled_ssq * n) as i32 - (scaled_sum * scaled_sum) as i32)
-      as u32;
+  let p = (scaled_ssq * n).saturating_sub(scaled_sum * scaled_sum);
   let z = (p * s + (1 << SGRPROJ_MTABLE_BITS >> 1)) >> SGRPROJ_MTABLE_BITS;
   let a = if z >= 255 {
     256
@@ -359,14 +362,20 @@ fn sgrproj_sum_finish(
 }
 
 // Using an integral image, compute the sum of a square region
-const fn get_integral_square(
+// SAFETY: The size of `iimg` must be at least `(y + size) * stride + x + size`
+#[inline(always)]
+unsafe fn get_integral_square(
   iimg: &[u32], stride: usize, x: usize, y: usize, size: usize,
 ) -> u32 {
   // Cancel out overflow in iimg by using wrapping arithmetic
-  iimg[y * stride + x]
-    .wrapping_add(iimg[(y + size) * stride + x + size])
-    .wrapping_sub(iimg[(y + size) * stride + x])
-    .wrapping_sub(iimg[y * stride + x + size])
+  let top_left = *iimg.get_unchecked(y * stride + x);
+  let top_right = *iimg.get_unchecked(y * stride + x + size);
+  let bottom_left = *iimg.get_unchecked((y + size) * stride + x);
+  let bottom_right = *iimg.get_unchecked((y + size) * stride + x + size);
+  top_left
+    .wrapping_add(bottom_right)
+    .wrapping_sub(bottom_left)
+    .wrapping_sub(top_right)
 }
 
 struct VertPaddedIter<'a, T: Pixel> {
@@ -517,6 +526,7 @@ impl<'a, T: Pixel> Iterator for HorzPaddedIter<'a, T> {
 impl<T: Pixel> ExactSizeIterator for HorzPaddedIter<'_, T> {}
 impl<T: Pixel> FusedIterator for HorzPaddedIter<'_, T> {}
 
+#[hawktracer(setup_integral_image)]
 pub fn setup_integral_image<T: Pixel>(
   integral_image_buffer: &mut IntegralImageBuffer,
   integral_image_stride: usize, crop_w: usize, crop_h: usize, stripe_w: usize,
@@ -616,13 +626,13 @@ pub fn setup_integral_image<T: Pixel>(
   }
 }
 
+#[hawktracer(sgrproj_stripe_filter)]
 pub fn sgrproj_stripe_filter<T: Pixel, U: Pixel>(
   set: u8, xqd: [i8; 2], fi: &FrameInvariants<T>,
   integral_image_buffer: &IntegralImageBuffer, integral_image_stride: usize,
   cdeffed: &PlaneSlice<U>, out: &mut PlaneRegionMut<U>,
 ) {
   let &Rect { width: stripe_w, height: stripe_h, .. } = out.rect();
-  let bdm8 = fi.sequence.bit_depth - 8;
   let mut a_r2: [[u32; IMAGE_WIDTH_MAX + 2]; 2] =
     [[0; IMAGE_WIDTH_MAX + 2]; 2];
   let mut b_r2: [[u32; IMAGE_WIDTH_MAX + 2]; 2] =
@@ -638,6 +648,19 @@ pub fn sgrproj_stripe_filter<T: Pixel, U: Pixel>(
   let s_r2: u32 = SGRPROJ_PARAMS_S[set as usize][0];
   let s_r1: u32 = SGRPROJ_PARAMS_S[set as usize][1];
 
+  let fn_ab_r1 = match fi.sequence.bit_depth {
+    8 => sgrproj_box_ab_r1::<8>,
+    10 => sgrproj_box_ab_r1::<10>,
+    12 => sgrproj_box_ab_r1::<12>,
+    _ => unimplemented!(),
+  };
+  let fn_ab_r2 = match fi.sequence.bit_depth {
+    8 => sgrproj_box_ab_r2::<8>,
+    10 => sgrproj_box_ab_r2::<10>,
+    12 => sgrproj_box_ab_r2::<12>,
+    _ => unimplemented!(),
+  };
+
   /* prime the intermediate arrays */
   // One oddness about the radius=2 intermediate array computations that
   // the spec doesn't make clear: Although the spec defines computation
@@ -646,7 +669,7 @@ pub fn sgrproj_stripe_filter<T: Pixel, U: Pixel>(
   let integral_image = &integral_image_buffer.integral_image;
   let sq_integral_image = &integral_image_buffer.sq_integral_image;
   if s_r2 > 0 {
-    sgrproj_box_ab_r2(
+    fn_ab_r2(
       &mut a_r2[0],
       &mut b_r2[0],
       integral_image,
@@ -655,13 +678,12 @@ pub fn sgrproj_stripe_filter<T: Pixel, U: Pixel>(
       0,
       stripe_w,
       s_r2,
-      bdm8,
       fi.cpu_feature_level,
     );
   }
   if s_r1 > 0 {
     let integral_image_offset = integral_image_stride + 1;
-    sgrproj_box_ab_r1(
+    fn_ab_r1(
       &mut a_r1[0],
       &mut b_r1[0],
       &integral_image[integral_image_offset..],
@@ -670,10 +692,9 @@ pub fn sgrproj_stripe_filter<T: Pixel, U: Pixel>(
       0,
       stripe_w,
       s_r1,
-      bdm8,
       fi.cpu_feature_level,
     );
-    sgrproj_box_ab_r1(
+    fn_ab_r1(
       &mut a_r1[1],
       &mut b_r1[1],
       &integral_image[integral_image_offset..],
@@ -682,7 +703,6 @@ pub fn sgrproj_stripe_filter<T: Pixel, U: Pixel>(
       1,
       stripe_w,
       s_r1,
-      bdm8,
       fi.cpu_feature_level,
     );
   }
@@ -693,7 +713,7 @@ pub fn sgrproj_stripe_filter<T: Pixel, U: Pixel>(
   for y in (0..stripe_h).step_by(2) {
     // get results to use y and y+1
     let f_r2_ab: [&[u32]; 2] = if s_r2 > 0 {
-      sgrproj_box_ab_r2(
+      fn_ab_r2(
         &mut a_r2[(y / 2 + 1) % 2],
         &mut b_r2[(y / 2 + 1) % 2],
         integral_image,
@@ -702,7 +722,6 @@ pub fn sgrproj_stripe_filter<T: Pixel, U: Pixel>(
         y + 2,
         stripe_w,
         s_r2,
-        bdm8,
         fi.cpu_feature_level,
       );
       let ap0: [&[u32]; 2] = [&a_r2[(y / 2) % 2], &a_r2[(y / 2 + 1) % 2]];
@@ -733,7 +752,7 @@ pub fn sgrproj_stripe_filter<T: Pixel, U: Pixel>(
       let y = y + dy;
       if s_r1 > 0 {
         let integral_image_offset = integral_image_stride + 1;
-        sgrproj_box_ab_r1(
+        fn_ab_r1(
           &mut a_r1[(y + 2) % 3],
           &mut b_r1[(y + 2) % 3],
           &integral_image[integral_image_offset..],
@@ -742,7 +761,6 @@ pub fn sgrproj_stripe_filter<T: Pixel, U: Pixel>(
           y + 2,
           stripe_w,
           s_r1,
-          bdm8,
           fi.cpu_feature_level,
         );
         let ap1: [&[u32]; 3] =
@@ -825,13 +843,12 @@ pub fn sgrproj_stripe_filter<T: Pixel, U: Pixel>(
 
 // Input params follow the same rules as sgrproj_stripe_filter.
 // Inputs are relative to the colocated slice views.
+#[hawktracer(sgrproj_solve)]
 pub fn sgrproj_solve<T: Pixel>(
   set: u8, fi: &FrameInvariants<T>,
   integral_image_buffer: &IntegralImageBuffer, input: &PlaneRegion<'_, T>,
   cdeffed: &PlaneSlice<T>, cdef_w: usize, cdef_h: usize,
 ) -> (i8, i8) {
-  let bdm8 = fi.sequence.bit_depth - 8;
-
   let mut a_r2: [[u32; IMAGE_WIDTH_MAX + 2]; 2] =
     [[0; IMAGE_WIDTH_MAX + 2]; 2];
   let mut b_r2: [[u32; IMAGE_WIDTH_MAX + 2]; 2] =
@@ -850,6 +867,19 @@ pub fn sgrproj_solve<T: Pixel>(
   let mut h: [[f64; 2]; 2] = [[0., 0.], [0., 0.]];
   let mut c: [f64; 2] = [0., 0.];
 
+  let fn_ab_r1 = match fi.sequence.bit_depth {
+    8 => sgrproj_box_ab_r1::<8>,
+    10 => sgrproj_box_ab_r1::<10>,
+    12 => sgrproj_box_ab_r1::<12>,
+    _ => unimplemented!(),
+  };
+  let fn_ab_r2 = match fi.sequence.bit_depth {
+    8 => sgrproj_box_ab_r2::<8>,
+    10 => sgrproj_box_ab_r2::<10>,
+    12 => sgrproj_box_ab_r2::<12>,
+    _ => unimplemented!(),
+  };
+
   /* prime the intermediate arrays */
   // One oddness about the radius=2 intermediate array computations that
   // the spec doesn't make clear: Although the spec defines computation
@@ -858,7 +888,7 @@ pub fn sgrproj_solve<T: Pixel>(
   let integral_image = &integral_image_buffer.integral_image;
   let sq_integral_image = &integral_image_buffer.sq_integral_image;
   if s_r2 > 0 {
-    sgrproj_box_ab_r2(
+    fn_ab_r2(
       &mut a_r2[0],
       &mut b_r2[0],
       integral_image,
@@ -867,13 +897,12 @@ pub fn sgrproj_solve<T: Pixel>(
       0,
       cdef_w,
       s_r2,
-      bdm8,
       fi.cpu_feature_level,
     );
   }
   if s_r1 > 0 {
     let integral_image_offset = SOLVE_IMAGE_STRIDE + 1;
-    sgrproj_box_ab_r1(
+    fn_ab_r1(
       &mut a_r1[0],
       &mut b_r1[0],
       &integral_image[integral_image_offset..],
@@ -882,10 +911,9 @@ pub fn sgrproj_solve<T: Pixel>(
       0,
       cdef_w,
       s_r1,
-      bdm8,
       fi.cpu_feature_level,
     );
-    sgrproj_box_ab_r1(
+    fn_ab_r1(
       &mut a_r1[1],
       &mut b_r1[1],
       &integral_image[integral_image_offset..],
@@ -894,7 +922,6 @@ pub fn sgrproj_solve<T: Pixel>(
       1,
       cdef_w,
       s_r1,
-      bdm8,
       fi.cpu_feature_level,
     );
   }
@@ -905,7 +932,7 @@ pub fn sgrproj_solve<T: Pixel>(
   for y in (0..cdef_h).step_by(2) {
     // get results to use y and y+1
     let f_r2_01: [&[u32]; 2] = if s_r2 > 0 {
-      sgrproj_box_ab_r2(
+      fn_ab_r2(
         &mut a_r2[(y / 2 + 1) % 2],
         &mut b_r2[(y / 2 + 1) % 2],
         integral_image,
@@ -914,7 +941,6 @@ pub fn sgrproj_solve<T: Pixel>(
         y + 2,
         cdef_w,
         s_r2,
-        bdm8,
         fi.cpu_feature_level,
       );
       let ap0: [&[u32]; 2] = [&a_r2[(y / 2) % 2], &a_r2[(y / 2 + 1) % 2]];
@@ -939,7 +965,7 @@ pub fn sgrproj_solve<T: Pixel>(
       let y = y + dy;
       if s_r1 > 0 {
         let integral_image_offset = SOLVE_IMAGE_STRIDE + 1;
-        sgrproj_box_ab_r1(
+        fn_ab_r1(
           &mut a_r1[(y + 2) % 3],
           &mut b_r1[(y + 2) % 3],
           &integral_image[integral_image_offset..],
@@ -948,7 +974,6 @@ pub fn sgrproj_solve<T: Pixel>(
           y + 2,
           cdef_w,
           s_r1,
-          bdm8,
           fi.cpu_feature_level,
         );
         let ap1: [&[u32]; 3] =
@@ -1070,6 +1095,7 @@ pub fn sgrproj_solve<T: Pixel>(
   }
 }
 
+#[hawktracer(wiener_stripe_filter)]
 fn wiener_stripe_filter<T: Pixel>(
   coeffs: [[i8; 3]; 2], fi: &FrameInvariants<T>, crop_w: usize, crop_h: usize,
   stripe_w: usize, stripe_h: usize, stripe_x: usize, stripe_y: isize,
